import RPi.GPIO as GPIO
from pi_awning_webthing.awning import Motor
from dataclasses import dataclass
from typing import List
import logging
from os import path


@dataclass
class Config:
    name: str
    gpio_forward: int
    gpio_backward: int


def load_tb6612fng(filename: str) -> List[Motor]:
    motors = list()
    if "tb6612fng" in filename.lower() and path.exists(filename):
        with open(filename, "r") as file:
            for line in file.readlines():
                line = line.strip()
                if not line.startswith("#") and len(line) > 0:
                    try:
                        parts = line.split(",")
                        name = parts[0].strip()
                        pin_forward = int(parts[1].strip())
                        pin_backward = int(parts[2].strip())
                        step_duration = float(parts[3].strip())
                        motors.append(TB6612FNGMotor(name, pin_forward, pin_backward, step_duration))
                        logging.info("config entry found: " + name + " with pin_forward=" + str(pin_forward) + ", pin_backward=" + str(pin_backward) + ", step_duration=" + str(step_duration))
                    except Exception as e:
                        logging.error("invalid syntax in line " + line + "  ignoring it" + str(e))
    return motors



class TB6612FNGMotor(Motor):

    def __init__(self, name: str, pin_forward: int, pin_backward: int, sec_per_step: float):
        self.__name = name
        self.__sec_per_step = sec_per_step
        GPIO.setmode(GPIO.BCM)
        self.pin_forward = pin_forward
        self.pin_forward_is_on = False
        GPIO.setup(pin_forward, GPIO.OUT, initial=0)
        self.pin_backward = pin_backward
        self.pin_backward_is_on = False
        GPIO.setup(pin_backward, GPIO.OUT, initial=0)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def sec_per_step(self) -> float:
        return self.__sec_per_step


    def stop(self):
        if self.pin_backward_is_on or self.pin_forward_is_on:
            logging.info("stop motor")
        if self.pin_backward_is_on:
            GPIO.output(self.pin_backward, 0)
            self.pin_backward_is_on = False
        if self.pin_forward_is_on:
            GPIO.output(self.pin_forward, 0)
            self.pin_forward_is_on = False

    def backward(self):
        self.stop()
        logging.info("start backward motor")
        GPIO.output(self.pin_backward, 1)
        self.pin_backward_is_on = True

    def forward(self):
        self.stop()
        logging.info("start forward motor")
        GPIO.output(self.pin_forward, 1)
        self.pin_forward_is_on = True




import logging
import sys
import time
import schedule
from datetime import datetime
from abc import ABC, abstractmethod
from threading import Thread


class Motor(ABC):

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def backward(self):
        pass

    @abstractmethod
    def forward(self):
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def sec_per_step(self) -> float:
        pass


class AwningPropertyListener:

    def on_current_pos_updated(self, current_position: int):
        pass

    def on_retracting_updated(self, retracting: bool):
        pass

    def on_extenting_updated(self, extenting: bool):
        pass


class Movement:
    SLOT_TOLERANCE = 7

    def __init__(self, motor: Motor, start_pos: int, num_slots: int, sec_per_slot: float, is_positive: bool, awning):
        self.start_time = datetime.now()
        self.awning = awning
        self.motor = motor
        self.start_pos = start_pos
        self.num_slots = num_slots
        self.sec_per_slot = sec_per_slot
        if is_positive:
            self.direction = 1
        else:
            self.direction = -1

    def get_pause_sec(self):
        return 0.5

    def get_current_pos(self) -> int:
        if self.is_target_reached():
            return self.get_target_pos()
        else:
            return self.start_pos + (self.__get_num_processed_slots() * self.direction)

    def get_target_pos(self) -> int:
        return self.start_pos + (self.num_slots * self.direction)

    def is_target_reached(self) -> bool:
        return self.__get_num_processed_slots() >= self.num_slots

    def __get_num_processed_slots(self) -> int:
        elapsed_sec = (datetime.now() - self.start_time).total_seconds()
        num_processed = 0
        if elapsed_sec > 1:
            num_processed = elapsed_sec / self.sec_per_slot
        return int(num_processed)

    def process(self):
        if self.is_target_reached():
            return Idling(self.motor, self.get_target_pos(), self.sec_per_slot, self.awning)
        else:
            self.awning.listener.on_current_pos_updated(self.get_current_pos())
            return self

    def drive_to(self, new_position: int):
        if new_position > 100:
            new_position = 100
        elif new_position < 0:
            new_position = 0
        return self.__create_movement(int(new_position))

    def __create_movement(self, new_position: int):
        current_pos = self.get_current_pos()
        if (new_position - current_pos) > self.SLOT_TOLERANCE:
            return Forward(self.motor, current_pos, new_position, self.sec_per_slot, self.awning)
        elif (current_pos - new_position) > self.SLOT_TOLERANCE:
            return Backward(self.motor, current_pos, new_position, self.sec_per_slot, self.awning)
        else:
            return Idling(self.motor, current_pos, self.sec_per_slot, self.awning)


class Idling(Movement):

    def __init__(self, motor: Motor, start_pos: int, sec_per_slot: float, awning):
        Movement.__init__(self, motor, start_pos, 0, sec_per_slot, True, awning)
        self.motor.stop()
        self.awning.listener.on_extenting_updated(False)
        self.awning.listener.on_retracting_updated(False)

    def get_pause_sec(self):
        pause_sec = int(self.SLOT_TOLERANCE * self.sec_per_slot * 1.4)
        if pause_sec < 3:
            pause_sec = 3
        return pause_sec

    def process(self):
        return self   # do nothing


class Forward(Movement):

    def __init__(self, motor: Motor, start_pos: int, new_position: int, sec_per_slot: float, awning):
        Movement.__init__(self, motor, start_pos, new_position - start_pos, sec_per_slot, True, awning)
        self.motor.forward()
        self.awning.listener.on_extenting_updated(True)
        self.awning.listener.on_retracting_updated(False)


class Backward(Movement):

    def __init__(self, motor: Motor, start_pos: int, new_position: int, sec_per_slot: float, awning):
        Movement.__init__(self, motor, start_pos, start_pos - new_position, sec_per_slot, False, awning)
        self.motor.backward()
        self.awning.listener.on_retracting_updated(True)
        self.awning.listener.on_extenting_updated(False)


class Awning:

    def __init__(self, motor: Motor):
        self.name = motor.name
        self.sec_per_slot = motor.sec_per_step
        self.logger = logging.getLogger(self.name)
        self.listener = AwningPropertyListener()
        self.motor = motor
        self.movement = Idling(self.motor, 0, self.sec_per_slot, self)
        Thread(name=self.name + "_move", target=self.__process_move, daemon=False).start()
        schedule.every().day.at("04:40").do(self.calibrate)
        schedule.run_pending()

    def register_listener(self, listener: AwningPropertyListener):
        self.listener = listener

    def calibrate(self):
        self.logger.info("calibrating")
        self.movement = Idling(self.motor, 100, self.sec_per_slot, self) # set position to 100%
        self.set_target_position(0)   # and backward to position 0. This ensures that the awning is calibrated with position 0

    def get_current_position(self) -> int:
        return self.movement.get_current_pos()

    def get_target_position(self) -> int:
        return self.movement.get_target_pos()

    def set_target_position(self, new_position: int):
        self.movement = self.movement.drive_to(new_position)

    def __process_move(self):
        time.sleep(1)
        self.calibrate()
        while True:
            try:
                self.movement = self.movement.process()
            except:
                self.movement = Idling(self.motor, 0, self.sec_per_slot, self)
                self.logger.warning('move operation failed '  + str(sys.exc_info()))
            finally:
                pause_sec = self.movement.get_pause_sec()
                time.sleep(pause_sec)



import logging
import RPi.GPIO as GPIO
from typing import List
from pi_awning_webthing.awning import Awning


class Switch:
    STOP = (False, False)
    MOVE_FORWARD = (True, False)
    MOVE_BACKWARD = (False, True)
    IDLE = (True, True)

    def __init__(self, pin_forward: int, pin_backward: int, awnings: List[Awning]):
        self.awnings = awnings
        self.pin_forward = pin_forward
        self.pin_backward = pin_backward
        self.state = self.STOP
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.pin_forward, GPIO.IN, GPIO.PUD_DOWN)
        GPIO.add_event_detect(self.pin_forward, GPIO.BOTH)
        GPIO.add_event_callback(self.pin_forward, self.on_switch_updated)
        GPIO.setup(self.pin_backward, GPIO.IN, GPIO.PUD_DOWN)
        GPIO.add_event_detect(self.pin_backward, GPIO.BOTH)
        GPIO.add_event_callback(self.pin_backward, self.on_switch_updated)
        logging.info("Switch bound to pin_forward=" + str(self.pin_forward) + " and pin_backward=" + str(self.pin_backward))

    def is_forward(self) -> bool:
        return self.state[0]

    def is_backward(self) -> bool:
        return self.state[1]

    def on_switch_updated(self, pin: int):
        is_forward = GPIO.input(self.pin_forward) >= 1
        is_backward = GPIO.input(self.pin_backward) >= 1
        new_state = (is_forward, is_backward)
        if new_state != self.state:
            logging.info("switch new state     " + str(new_state) + " (old " + str(self.state) + ")")
        try:
            if self.state != new_state:
                if new_state == self.MOVE_FORWARD:
                    for anwing in self.awnings:
                        anwing.set_target_position(100)
                elif new_state == self.MOVE_BACKWARD:
                    for anwing in self.awnings:
                        anwing.set_target_position(0)
                elif new_state == self.STOP:
                    for anwing in self.awnings:
                        current_pos = anwing.get_current_position()
                        anwing.set_target_position(current_pos)
                else:
                    pass
        finally:
            self.state = new_state



from webthing import (MultipleThings, Property, Thing, Value, WebThingServer)
from pi_awning_webthing.awning import Awning, AwningPropertyListener
from pi_awning_webthing.switch import Switch
from pi_awning_webthing.motor_tb6612Fng import load_tb6612fng
import logging
import tornado.ioloop


class WebThingAwningPropertyListener(AwningPropertyListener):

    def __init__(self, anwing_webthing):
        self.anwing_webthing = anwing_webthing

    def on_current_pos_updated(self, current_position: int):
        self.anwing_webthing.ioloop.add_callback(self.anwing_webthing.set_current_position, current_position)

    def on_retracting_updated(self, retracting: bool):
        self.anwing_webthing.ioloop.add_callback(self.anwing_webthing.set_retracting, retracting)

    def on_extenting_updated(self, extenting: bool):
        self.anwing_webthing.ioloop.add_callback(self.anwing_webthing.set_extending, extenting)


class AnwingWebThing(Thing):

    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing

    def __init__(self, description: str, awning: Awning):
        Thing.__init__(
            self,
            'urn:dev:ops:anwing-TB6612FNG',
            'Awning ' + awning.name + " Controller",
            ['MultiLevelSensor'],
            description
        )
        self.logger = logging.getLogger(awning.name)
        self.awning = awning
        self.awning.register_listener(WebThingAwningPropertyListener(self))

        self.target_position = Value(0, self.__target_position)
        self.add_property(
            Property(self,
                     'target_position',
                     self.target_position,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Awning target position',
                         "type": "integer",
                         "minimum": 0,
                         "maximum": 100,
                         "unit": "percent",
                         'description': 'awning target position'
                     }))

        self.current_position = Value(0)
        self.add_property(
            Property(self,
                     'current_position',
                     self.current_position,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Awning current position',
                         "type": "integer",
                         'minimum': 0,
                         'maximum': 100,
                         "unit": "percent",
                         'readOnly': True,
                         'description': 'awning current position'
                     }))

        self.retracting = Value(0)
        self.add_property(
        Property(self,
                 'retracting',
                 self.retracting,
                 metadata={
                     '@type': 'BooleanProperty',
                     'title': 'Awning is retracting',
                     "type": "boolean",
                     'readOnly': True,
                     'description': 'Awning is retracting'
                 }))

        self.extending = Value(0)
        self.add_property(
        Property(self,
                 'extending',
                 self.extending,
                 metadata={
                     '@type': 'BooleanProperty',
                     'title': 'Awning is extending',
                     "type": "boolean",
                     'readOnly': True,
                     'description': 'Awning is extending'
                 }))

        self.ioloop = tornado.ioloop.IOLoop.current()

    def __target_position(self, new_postion):
        self.awning.set_target_position(new_postion)

    def set_current_position(self, value):
        self.current_position.notify_of_external_update(value)
        self.logger.info("position " + str(value) + " reached (target=" + str(self.target_position.get()) + ")")

    def set_retracting(self, value):
        self.retracting.notify_of_external_update(value)
        self.logger.info("awning is retracting " + str(value))

    def set_extending(self, value):
        self.extending.notify_of_external_update(value)
        self.logger.info("awning is extending " + str(value))



def run_server(port: int, filename: str, switch_pin_forward: int, switch_pin_backward: int, description: str):
    awnings = [Awning(motor) for motor in load_tb6612fng(filename)]
    awning_webthings = [AnwingWebThing(description, anwing) for anwing in awnings]
    server = WebThingServer(MultipleThings(awning_webthings, 'Awnings'), port=port, disable_host_validation=True)
    if switch_pin_forward > 0 and switch_pin_backward > 0:
        Switch(switch_pin_forward, switch_pin_backward, awnings= awnings)

    try:
        logging.info('starting the server')
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping the server')
        server.stop()
        logging.info('done')



import logging
from string import Template
from pi_awning_webthing.app import App
from pi_awning_webthing.awning_webthing import run_server

PACKAGENAME = 'pi_awning_webthing'
ENTRY_POINT = "awning"
DESCRIPTION = "A web connected terrace awning controller on Raspberry Pi"


UNIT_TEMPLATE = Template('''
[Unit]
Description=$packagename
After=syslog.target

[Service]
Type=simple
ExecStart=$entrypoint --command listen --verbose $verbose --port $port --filename $filename --switch_pin_forward $switch_pin_forward --switch_pin_backward $switch_pin_backward
SyslogIdentifier=$packagename
StandardOutput=syslog
StandardError=syslog
Restart=always
RestartSec=3

[Install]
WantedBy=multi-user.target
''')


class AwningApp(App):

    def do_add_argument(self, parser):
        parser.add_argument('--filename', metavar='filename', required=False, type=str,  help='the config filename')
        parser.add_argument('--switch_pin_forward', metavar='switch_pin_forward', required=False, type=int, default=-1, help='foward pin for switch')
        parser.add_argument('--switch_pin_backward', metavar='switch_pin_backward', required=False, type=int, default=-1, help='backward pin for switch')

    def do_additional_listen_example_params(self):
        return "--filename /etc/awning/tb6612fng_motors.config --switch_pin_forward 17 --switch_pin_backward 27"

    def do_process_command(self, command:str, port: int, verbose: bool, args) -> bool:
        if command == 'listen' and (args.filename is not None):
            logging.info("running " + self.packagename + " on port " + str(port) + " with config " + args.filename)
            run_server(port, args.filename, args.switch_pin_forward, args.switch_pin_backward, self.description)
            return True
        elif args.command == 'register' and (args.filename is not None):
            logging.info("register " + self.packagename + " on port " + str(port) + " with config " + args.filename)
            unit = UNIT_TEMPLATE.substitute(packagename=self.packagename, entrypoint=self.entrypoint, port=port, verbose=verbose, switch_pin_forward=args.switch_pin_forward, switch_pin_backward=args.switch_pin_backward, filename=args.filename)
            self.unit.register(port, unit)
            return True
        else:
            return False


def main():
    AwningApp(PACKAGENAME, ENTRY_POINT, DESCRIPTION).handle_command()


if __name__ == '__main__':
    main()
